import { Component, Input, Output, EventEmitter, ViewChildren, QueryList, DoCheck } from "@angular/core";

import { ParamRadioConfig } from "../../formulaire/ngparam";
import { FieldSet } from "../../formulaire/fieldset";
import { ParamFieldLineComponent } from "../param-field-line/param-field-line.component";
import { Field } from "../../formulaire/field";
import { InputField } from "../../formulaire/input-field";
import { SelectField } from "../../formulaire/select-field";
import { CheckField } from "../../formulaire/check-field";

@Component({
    selector: "field-set",
    templateUrl: "./field-set.component.html",
    styleUrls: ["./field-set.component.scss"]
})
export class FieldSetComponent implements DoCheck {

    @Input()
    private set fieldSet(fs: FieldSet) {
        this._fieldSet = fs;
    }

    public get fields() {
        return this._fieldSet.kids;
    }

    public get showButtons() {
        return this._showButtons;
    }

    public set showButtons(b: boolean) {
        this._showButtons = b;
    }

    public set enableUpButton(b: boolean) {
        this._enableUpButton = b;
    }

    public set enableDownButton(b: boolean) {
        this._enableDownButton = b;
    }

    public set enableRemoveButton(b: boolean) {
        this._enableRemoveButton = b;
    }

    public set fieldsetNumber(n: number) {
        this._fieldSet.labelNumber = n;
    }

    public get title(): string {
        if (! this._fieldSet) {
            return "fs undefined";
        }
        if (! this._fieldSet.label) { // @TODO allow "" ?
            return "label undefined";
        }
        return this._fieldSet.label;
    }

    public get isValid() {
        return this._isValid;
    }

    /**
     * couleur du bouton monter
     */
    private get upButtonColor(): string {
        return this._enableUpButton ? "black" : "lightgrey";
    }

    /**
     * couleur du bouton descendre
     */
    private get downButtonColor(): string {
        return this._enableDownButton ? "black" : "lightgrey";
    }

    /**
     * couleur du bouton supprimer
     */
    private get removeButtonColor(): string {
        return this._enableRemoveButton ? "black" : "lightgrey";
    }
    /**
    * field set attribute
    */
    private _fieldSet: FieldSet;

    @ViewChildren(ParamFieldLineComponent)
    private _paramComponents: QueryList<ParamFieldLineComponent>;

    /**
     * événement de changement de validité
     */
    @Output()
    private validChange = new EventEmitter();

    /**
     * événement de changement de valeur d'un input
     */
    @Output()
    private inputChange = new EventEmitter();

    /**
     * événement de demande d'ajout d'un fieldset (FieldSet dans un FieldsetContainer)
     */
    @Output()
    private addFieldset = new EventEmitter();

    /**
     * événement de demande de remontée d'un fieldset dans la liste (FieldSet dans un FieldsetContainer)
     */
    @Output()
    private moveFieldsetUp = new EventEmitter();

    /**
     * événement de demande de descente d'un fieldset dans la liste (FieldSet dans un FieldsetContainer)
     */
    @Output()
    private moveFieldsetDown = new EventEmitter();

    /**
     * événement de demande de suppression d'un fieldset (FieldSet dans un FieldsetContainer)
     */
    @Output()
    private removeFieldset = new EventEmitter();

    /**
     * flag de validité de la saisie
     */
    private _isValid = false;

    /**
     * flag d'affichage des boutons ajouter, supprimer, monter, descendre
     */
    private _showButtons = false;

    /**
     * flag d'activation du bouton monter
     */
    private _enableUpButton = true;

    /**
     * flag d'activation du bouton descendre
     */
    private _enableDownButton = true;

    /**
     * flag d'activation du bouton supprimer
     */
    private _enableRemoveButton = true;

    /**
     * événement de changement d'état d'un radio
     */
    // tslint:disable-next-line:no-output-on-prefix
    @Output()
    private radio = new EventEmitter<any>();

    private hasRadioFix(): boolean {
        if (this._fieldSet.hasInputs) {
            switch (this._fieldSet.getInput(0).radioConfig) {
                case ParamRadioConfig.FIX:
                    return false;

                default:
                    return true;
            }
        }
        return false;
    }

    private hasRadioVar(): boolean {
        if (this._fieldSet.hasInputs) {
            switch (this._fieldSet.getInput(0).radioConfig) {
                case ParamRadioConfig.VAR:
                case ParamRadioConfig.CAL:
                    return true;

                default:
                    return false;
            }
        }
        return false;
    }

    private hasRadioCal(): boolean {
        if (this._fieldSet.hasInputs) {
            switch (this._fieldSet.getInput(0).radioConfig) {
                case ParamRadioConfig.CAL:
                    return true;

                default:
                    return false;
            }
        }
        return false;
    }

    /**
     * détermine si un Field est du type InputField
     */
    private isInputField(f: Field): boolean {
        return f instanceof InputField && f.isDisplayed;
    }

    /**
     * détermine si un Field est du type SelectField
     */
    private isSelectField(f: Field): boolean {
        return f instanceof SelectField && f.isDisplayed;
    }

    /**
     * détermine si un Field est du type CheckField
     */
    private isCheckField(f: Field): boolean {
        return f instanceof CheckField && f.isDisplayed;
    }

    /*
     * gestion des événements clic sur les radios :
     * réception d'un message du composant enfant (param-field)
     * cf. https://angular.io/guide/component-interaction#parent-listens-for-child-event
     */
    private onRadioClick(info: any) {
        // on renvoie l'info au parent
        this.radio.emit(info);
    }

    /**
     * calcul de la validité de tous les ParamFieldLineComponent de la vue
     */
    private updateValidity() {
        this._isValid = false;

        if (this._paramComponents) {
            this._isValid = this._paramComponents.reduce(
                // callback
                (
                    // accumulator (valeur précédente du résultat)
                    acc,
                    // currentValue (élément courant dans le tableau)
                    param,
                    // currentIndex (indice courant dans le tableau)
                    currIndex,
                    // array (tableau parcouru)
                    array
                ) => {
                    return acc && param.isValid;
                }
                // valeur initiale
                , true);
        }

        this.validChange.emit();
    }

    public ngDoCheck() {
        this.updateValidity();
    }

    /**
     * réception d'un événement de validité de ParamFieldLineComponent
     */
    private onParamLineValid(event: boolean) {
        this.updateValidity();
    }

    /**
    * réception d'un événement de changement de valeur d'un input
    */
    private onInputChange(event: boolean) {
        this.inputChange.emit();
    }

    /**
     * relit les valeurs dans l'interface et met à jour les NgParameter
     */
    public updateParametersFromUI() {
        this._paramComponents.forEach(fsc => fsc.updateParameterFromUI());
    }

    /**
     * met à jour les paramètres liés
     */
    public updateLinkedParameters() {
        this._paramComponents.forEach(fsc => fsc.updateLinkedParameter());
    }

    /**
     * clic sur le bouton ajouter
     */
    private onAddClick() {
        this.addFieldset.emit(this._fieldSet);
    }

    /**
     * clic sur le bouton supprimer
     */
    private onRemoveClick() {
        if (this._enableRemoveButton) {
            this.removeFieldset.emit(this._fieldSet);
        }
    }

    /**
     * clic sur le bouton monter
     */
    private onMoveUpClick() {
        if (this._enableUpButton) {
            this.moveFieldsetUp.emit(this._fieldSet);
        }
    }

    /**
     * clic sur le bouton descendre
     */
    private onMoveDownClick() {
        if (this._enableDownButton) {
            this.moveFieldsetDown.emit(this._fieldSet);
        }
    }
}
