import { FormulaireNode } from "./formulaire-node";
import { StringMap } from "../stringmap";
import { Dependency } from "./dependency/dependency";
import { DependencyCondition, DependencyConditionType } from "./dependency/dependency-condition";
import { ValueDependencyCondition } from "./dependency/value-dependency-condition";
import { ExistenceDependency } from "./dependency/existence-dependency";
import { DeepFormulaireElementIterator } from "./form-iterator/deep-element-iterator";
import { InternationalisationService } from "../services/internationalisation/internationalisation.service";
import { ServiceFactory } from "../services/service-factory";

/**
 * élément (enfant) du formulaire : fieldset, input, container, ...
 */
export abstract class FormulaireElement extends FormulaireNode {
    /**
     * affiché ?
     */
    private _isDisplayed: boolean;

    /**
     * étiquette affichée dans l'UI
     */
    private _label: string;

    /**
     * n° d'ordre ajouté à l'étiquette
     */
    private _labelNumber: number;

    private intlService: InternationalisationService;

    protected _dependencies: Dependency[] = [];

    public static removePrefix(s: string, prefix: string): string {
        if (s.startsWith(prefix)) {
            const l = prefix.length;
            return s.substr(l, s.length - l);
        }
        return undefined;
    }

    constructor(parent: FormulaireNode) {
        super(parent);
        this._isDisplayed = true;
        this.intlService = ServiceFactory.instance.internationalisationService;
    }

    get isDisplayed(): boolean {
        return this._isDisplayed;
    }

    set isDisplayed(b: boolean) {
        this._isDisplayed = b;
        for (const k of this.getKids()) {
            k.isDisplayed = b;
        }
    }

    get label(): string {
        let res = this._label;
        if (res && this._labelNumber) {
            res += " n°" + String(this._labelNumber);
        }
        return res;
    }

    set labelNumber(n: number) {
        this._labelNumber = n;
    }

    public getKids(): FormulaireElement[] {
        return this.kids as FormulaireElement[];
    }

    /**
     * formulaire parent
     */
    public get parentForm(): FormulaireNode {
        let res = this.parent;
        // while (!(res instanceof FormulaireDefinition))
        while (!("calculatorName" in res)) {
            // pour éviter de faire référence au type FormulaireDefinition, supprimer l'import correspondant et
            // casser les dépendances circulaires d'import
            res = res.parent;
        }
        return res;
    }

    /**
     * analyse les dépendances d'existence
     * @param json configuration de la dépendance
     */
    private parse_existence_dependencies(json: {}) {
        for (const di in json) {
            if (json.hasOwnProperty(di)) {
                const d = json[di];
                const rv = d["refvalue"];
                let mc: DependencyCondition;
                if (rv !== undefined) {
                    mc = new ValueDependencyCondition(rv);
                } else {
                    const cond = d["cond"];
                    if (cond !== undefined) {
                        switch (cond) {
                            case "isvar":
                                mc = new DependencyCondition(DependencyConditionType.IsVariable);
                                break;

                            case "isdisp":
                                mc = new DependencyCondition(DependencyConditionType.IsDisplayed);
                                break;

                            default:
                                throw new Error(
                                    "Formulaire.parse_existence_dependencies() : type de condition '" + cond + "' non pris en charge"
                                );
                        }
                    } else {
                        throw new Error("Formulaire.parse_existence_dependencies() : infos de dépendance manquantes/non prises en charge");
                    }
                }
                const dep = new ExistenceDependency(this, d["refid"], mc);
                this._dependencies.push(dep);
            }
        }
    }

    /**
     * analyse les dépendances (existence/valeur)
     * @param json configuration de la dépendance
     * @param parentNode node parent dans lequel rechercher l'élément master (dont l'existence dépend)
     */
    public parseDependencies(json: {}) {
        const dep = json["dep_exist"];
        if (dep !== undefined) {
            this.parse_existence_dependencies(dep);
        }
    }

    public verifiesDependency(d: Dependency): boolean {
        if (d.masterCondition.type === DependencyConditionType.IsDisplayed) {
            return this._isDisplayed;
        }
    }

    /**
     * applique une dépendance dont la condition a été vérifiée
     */
    public applyDependency(d: Dependency) {
        if (d instanceof ExistenceDependency) {
            this.isDisplayed = true;
        }
    }

    private prepareExistenceDependencies() {
        // si des FormulaireElement sont présents dans des ExistenceDependency, on met leur membre isDisplayed à false

        for (const d of this._dependencies) {
            if (d instanceof ExistenceDependency) {
                this.isDisplayed = false;
                break;
            }
        }
    }

    public applyDependencies() {
        this.prepareExistenceDependencies();

        for (const d of this._dependencies) {
            d.apply();
        }

        for (const k of this.getKids()) {
            k.applyDependencies();
        }
    }

    public printDependencies() {
        for (const d of this._dependencies) {
            console.log(d.toString());
        }
    }


    public updateLocalisation(loc: StringMap, key?: string) {
        if (!key) {
            key = this._confId;
        }
        if (loc[key] !== undefined) {
            this._label = this.intlService.localizeText(key, loc);
        } else {
            // Recherche du code dans locale/error_message.xx.json
            this._label = this.intlService.localizeText("INFO_LIB_" + key.replace(/\d+$/, "").toLocaleUpperCase());
        }
        for (const f of this.getKids()) {
            f.updateLocalisation(loc);
        }
    }

    public toString() {
        return "id:" + this.id + (this._isDisplayed ? " displayed" : " NOT displayed") + " label:" + this.label;
    }

    /**
     * itère sur tous les FormulaireElement
     */
    public get allFormElements(): IterableIterator<FormulaireElement> {
        return new DeepFormulaireElementIterator(this);
    }
}
