import { Interval, ParamDefinition, ParamDomain, ParamValueMode, INumberIterator, Nub, Observer, asObservable } from "jalhyd";

import { InputField } from "./input-field";
import { Dependency } from "./dependency/dependency";
import { DependencyConditionType } from "./dependency/dependency-condition";
import { ValueDependencyCondition } from "./dependency/value-dependency-condition";
import { ServiceFactory } from "../services/service-factory";
import { ApplicationSetupService } from "../services/app-setup/app-setup.service";
import { FormulaireNode } from "./formulaire-node";

export enum ParamRadioConfig {
    /**
     * pas de radio, paramètre modifiable à la main uniquement
     */
    FIX,

    /**
     * boutons radio "paramètre fixé" et "paramètre à varier"
     */
    VAR,

    /**
     * boutons radio "paramètre fixé", "paramètre à varier" et "paramètre à calculer"
     */
    CAL,

    /**
     * boutons radio "paramètre fixé", "paramètre à varier" et "paramètre à calculer", "paramètre lié"
     */
    LINK
}

/**
 * classe englobante de ParamDefinition (champs supplémentaires pour l'affichage, radio boutons, ...)
 */
export class NgParameter extends InputField implements Observer {

    constructor(private _paramDef: ParamDefinition, parent: FormulaireNode) {
        super(parent);
    }

    get symbol(): string {
        return this._paramDef.symbol;
    }

    get domain(): ParamDomain {
        return this._paramDef.getDomain();
    }

    public set confId(id: string) {
        if (this._confId !== undefined) {
            throw new Error(`NgParameter : l'identifiant de configuration est déjà défini (${this._confId})`);
        }
        this._confId = id;
    }

    public get _paramValues() { // @TODO remettre en private ! (debug)
        return this._paramDef.paramValues;
    }

    public get paramDefinition() {
        return this._paramDef;
    }

    public get radioState() {
        switch (this._paramDef.valueMode) {
            case ParamValueMode.SINGLE:
                return ParamRadioConfig.FIX;

            case ParamValueMode.LISTE:
            case ParamValueMode.MINMAX:
                return ParamRadioConfig.VAR;

            case ParamValueMode.CALCUL:
                return ParamRadioConfig.CAL;

            case ParamValueMode.LINK:
                return ParamRadioConfig.LINK;
        }
    }

    get isDefined(): boolean {
        return this._paramDef.isDefined;
    }

    public get valueMode() {
        return this._paramDef.valueMode;
    }

    public set valueMode(m: ParamValueMode) {
        // undefined si on clique en dehors du select après l'avoir ouvert (cad sans avoir fait de sélection)
        // et au même niveau, cad à côté du bouton et non à côté du menu déroulant
        if (m !== undefined && this._paramDef.valueMode !== m) {
            this.unlinkParameter();
            this._paramDef.valueMode = m;
            this.notifyObservers({
                "action": "valueModeChange",
                "value": this._paramDef.getValue()
            });
        }
    }

    public get isMinMaxValid(): boolean {
        return this._paramDef.isMinMaxValid;
    }

    public get minValue() {
        return this._paramValues.min;
    }

    public set minValue(v: number) {
        this._paramValues.min = v;
    }

    public get maxValue() {
        return this._paramValues.max;
    }

    public set maxValue(v: number) {
        this._paramValues.max = v;
    }

    public get stepRefValue(): Interval {
        return this._paramValues.stepRefValue;
    }

    public get stepValue() {
        return this._paramValues.step;
    }

    public set stepValue(v: number) {
        this._paramValues.step = v;
    }

    public get valueList() {
        return this._paramValues.valueList;
    }

    public set valueList(l: number[]) {
        this._paramValues.valueList = l;
    }

    public get isValid() {
        if (this.radioState === undefined) {
            return false;
        }

        return this._paramDef.isValid;
    }

    public get valuesIterator(): INumberIterator {
        return this._paramDef.valuesIterator;
    }
    public unit: string;
    public radioConfig: ParamRadioConfig;

    /**
     * true si ce paramètre est celui par défaut dans un formulaire (cf. fichier de conf des calculettes, objet "options", champ "idCal")
     */
    public isDefault = false; // archi bug du langage ! si on relit cette propriété sans l'avoir modifiée entre-temps, elle vaut undefined !

    private static getRadioConfig(s: string) {
        if (s === "fix") {
            return ParamRadioConfig.FIX;
        }

        if (s === "var") {
            return ParamRadioConfig.VAR;
        }

        if (s === "cal") {
            return ParamRadioConfig.CAL;
        }

        throw new Error("invalid parameter radio configuration " + s);
    }

    public getValue() {
        return this._paramDef.v;
    }

    /**
     * notification envoyée après la modification de la valeur du paramètre
     */
    private notifyValueModified(sender: any) {
        this.notifyObservers(
            {
                "action": "ngparamAfterValue",
                "param": this,
                "value": this._paramDef.v
            }, sender
        );
    }

    /**
     * fixe la valeur du paramètre.
     * une notification préalable est envoyée pour laisser l'occasion aux objets liés de préciser le contexte
     * dans lequel cette valeur existe
     * @param sender
     * @param val
     */
    public setValue(sender: any, val: number) {
        this._paramDef.setValue(val, sender);
        this.notifyValueModified(sender);
    }

    /**
     * supprime un lien avec un paramètre
     */
    private unlinkParameter() {
        const o = asObservable(this._paramDef.referencedObject);
        if (this.valueMode === ParamValueMode.LINK) {
            this._paramDef.undefineReference();
            if (o !== undefined) {
                o.removeObserver(this);
            }
        }
    }

    /**
     * crée le lien avec un paramètre
     */
    public linkToParameter(n: Nub, ref: string) {
        const changed: boolean =
            this._paramDef.valueMode !== ParamValueMode.LINK || this._paramDef.referencedNub !== n ||
            this._paramDef.referenceDefinition !== ref;
        if (changed) {
            let o = asObservable(this._paramDef.referencedObject);
            if (o !== undefined) {
                o.removeObserver(this);
            }

            this.valueMode = ParamValueMode.LINK;
            // changement de lien
            this._paramDef.defineReference(n, ref);

            o = asObservable(this._paramDef.referencedObject);
            if (o !== undefined) {
                o.addObserver(this); // pour être prévenu des changements de valeur de l'object référencé
            }

            this.notifyObservers({
                "action": "valueLinkChange",
                "value": this.getValue()
            });
        }
    }

    public checkValue(val: number) {
        this._paramDef.checkValue(val);
    }

    public checkList(l: number[]) {
        for (const e of l) {
            this.checkValue(e);
        }
    }

    public checkMin(min: number): boolean {
        return this._paramDef.checkMin(min);
    }

    public checkMax(max: number): boolean {
        return this._paramDef.checkMax(max);
    }

    public checkStep(step: number): boolean {
        return this._paramDef.checkStep(step);
    }

    public parseConfig(json: {}, data?: {}) {
        const appSetupService: ApplicationSetupService = ServiceFactory.instance.applicationSetupService;
        const radioConfig: string = data["radioConfig"];

        this._confId = json["id"];
        this.unit = json["unit"];
        let val: number;
        if (this.symbol === "Pr") {
            val = appSetupService.computePrecision;
        } else {
            val = json["value"];
        }
        if (val) {
            this.setValue(this, +val);
        }
        this.radioConfig = NgParameter.getRadioConfig(radioConfig);
        // tslint:disable-next-line:max-line-length
        this.isDefault = false; // malgré le fait qu'il soit initialisé dans la déclaration de la classe NgParam à false, quand on relit sa valeur, il vaut undefined (merci Microsoft)
    }

    public verifiesDependency(d: Dependency): boolean {
        if (super.verifiesDependency(d)) {
            return true;
        }

        switch (d.masterCondition.type) {
            case DependencyConditionType.HasValue: {
                    const mc: ValueDependencyCondition = <ValueDependencyCondition>d.masterCondition;
                    return this.getValue() === mc.value;
                }

            case DependencyConditionType.IsVariable:
                return this.radioState === ParamRadioConfig.VAR;

            default:
                throw new Error("NgParameter.verifiesDependency() : type de condition '"
                    + DependencyConditionType[d.masterCondition.type] + "' non pris en charge");
        }
    }

    private paramValuesJSON(): any {
        const res = {};
        const vm = this._paramDef.valueMode;
        res["mode"] = ParamValueMode[vm];
        switch (vm) {
            case ParamValueMode.SINGLE:
                res["value"] = this._paramValues.singleValue;
                break;

            case ParamValueMode.MINMAX:
                res["min"] = this._paramValues.min;
                res["max"] = this._paramValues.max;
                res["step"] = this._paramValues.step;
                break;

            case ParamValueMode.LISTE:
                res["values"] = this._paramValues.valueList;
                break;
        }
        return res;
    }

    /**
     * sérialisation en JSON
     */
    public JSONserialise(): {} {
        const res = {};
        res["id"] = this.id;
        res["values"] = this.paramValuesJSON();
        return { "param": res };
    }

    private deserialiseValues(json: {}) {
        const m: ParamValueMode = (<any>ParamValueMode)[json["mode"]];
        switch (m) {
            case ParamValueMode.SINGLE:
                this._paramValues.setValues(+json["value"]);
                break;

            case ParamValueMode.MINMAX:
                this._paramValues.setValues(+json["min"], +json["max"], +json["step"]);
                break;

            case ParamValueMode.LISTE:
                this._paramValues.setValues(json["values"]);
                break;

            case ParamValueMode.CALCUL:
                this._paramValues.valueMode = ParamValueMode.CALCUL;
                break;

            default:
                throw new Error(`session file : invalid value mode '${json["mode"]}' in param object`);
        }
    }

    /**
     * désérialisation depuis JSON
     */
    public deserialiseJSON(elements: {}) {
        for (const k in elements) {
            switch (k) {
                case "id":
                    break;

                case "values":
                    this.deserialiseValues(elements[k]);
                    break;

                default:
                    throw new Error(`session file : invalid key '${k}' in param object`);
            }
        }
    }

    // interface Observer

    public update(sender: any, data: any) {
        switch (data["action"]) {
            case "paramdefinitionAfterValue": // changement de valeur envoyé par l'objet référencé
                this.notifyValueModified(sender);
                break;
        }
    }
}
